from functools import partial
from unittest import (
    TestCase,
    mock,
)

from lxml import etree

from pcs.common.reports import ReportItemSeverity as severities
from pcs.common.reports import codes as report_codes
from pcs.common.tools import Version
from pcs.lib.cib import tools as lib

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_report_item_list_equal,
)
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.xml import get_xml_manipulation_creator_from_file

cib_element_lookup = etree.fromstring(
    """
            <cib>
                <configuration>
                    <resources>
                        <primitive id="R1" />
                        <primitive id="R3" />
                        <primitive id="RX2" />
                        <primitive id="RX2" />
                        <primitive id="RX3" />
                        <primitive id="T" />
                    </resources>
                    <tags>
                        <tag id="T">
                            <obj_ref id="RX1" />
                            <obj_ref id="RX3" />
                        </tag>
                    </tags>
                </configuration>
                <status>
                    <lrm_resource id="R3" />
                    <lrm_resource id="R2" />
                </status>
            </cib>
            """
)

# pylint: disable=no-self-use, line-too-long


class CibToolsTest(TestCase):
    def setUp(self):
        self.create_cib = get_xml_manipulation_creator_from_file(
            rc("cib-empty.xml")
        )
        self.cib = self.create_cib()

    def fixture_add_primitive_with_id(self, element_id):
        self.cib.append_to_first_tag_name(
            "resources",
            '<primitive id="{0}" class="ocf" provider="heartbeat" type="Dummy"/>'.format(
                element_id
            ),
        )


class IdProviderTest(CibToolsTest):
    def setUp(self):
        super().setUp()
        self.provider = lib.IdProvider(self.cib.tree)

    @staticmethod
    def fixture_report(_id):
        return (
            severities.ERROR,
            report_codes.ID_ALREADY_EXISTS,
            {
                "id": _id,
            },
            None,
        )


class IdProviderBook(IdProviderTest):
    def test_nonexisting_id(self):
        assert_report_item_list_equal(self.provider.book_ids("myId"), [])

    def test_existing_id(self):
        self.fixture_add_primitive_with_id("myId")
        assert_report_item_list_equal(
            self.provider.book_ids("myId"),
            [
                self.fixture_report("myId"),
            ],
        )

    def test_double_book(self):
        assert_report_item_list_equal(self.provider.book_ids("myId"), [])
        assert_report_item_list_equal(
            self.provider.book_ids("myId"),
            [
                self.fixture_report("myId"),
            ],
        )

    def test_more_ids(self):
        assert_report_item_list_equal(
            self.provider.book_ids("myId1", "myId2"), []
        )
        assert_report_item_list_equal(
            self.provider.book_ids("myId1", "myId2"),
            [
                self.fixture_report("myId1"),
                self.fixture_report("myId2"),
            ],
        )

    def test_complex(self):
        # test ids existing in the cib, double booked, available
        # test reports not repeated
        self.fixture_add_primitive_with_id("myId1")
        self.fixture_add_primitive_with_id("myId2")
        assert_report_item_list_equal(
            self.provider.book_ids(
                "myId1", "myId2", "myId3", "myId2", "myId3", "myId4", "myId3"
            ),
            [
                self.fixture_report("myId1"),
                self.fixture_report("myId2"),
                self.fixture_report("myId3"),
            ],
        )


class IdProviderAllocate(IdProviderTest):
    def test_nonexisting_id(self):
        self.assertEqual("myId", self.provider.allocate_id("myId"))

    def test_existing_id(self):
        self.fixture_add_primitive_with_id("myId")
        self.assertEqual("myId-1", self.provider.allocate_id("myId"))

    def test_allocate_books(self):
        self.assertEqual("myId", self.provider.allocate_id("myId"))
        self.assertEqual("myId-1", self.provider.allocate_id("myId"))

    def test_booked_ids(self):
        self.fixture_add_primitive_with_id("myId")
        assert_report_item_list_equal(self.provider.book_ids("myId-1"), [])
        self.assertEqual("myId-2", self.provider.allocate_id("myId"))


class DoesIdExistTest(CibToolsTest):
    def test_existing_id(self):
        self.fixture_add_primitive_with_id("myId")
        self.assertTrue(lib.does_id_exist(self.cib.tree, "myId"))

    def test_nonexisting_id(self):
        self.fixture_add_primitive_with_id("myId")
        self.assertFalse(lib.does_id_exist(self.cib.tree, "otherId"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "myid"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, " myId"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "myId "))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "my Id"))

    def test_ignore_status_section(self):
        self.cib.append_to_first_tag_name(
            "status",
            """
            <elem1 id="status-1">
                <elem1a id="status-1a">
                    <elem1aa id="status-1aa"/>
                    <elem1ab id="status-1ab"/>
                </elem1a>
                <elem1b id="status-1b">
                    <elem1ba id="status-1ba"/>
                    <elem1bb id="status-1bb"/>
                </elem1b>
            </elem1>
        """,
        )
        self.assertFalse(lib.does_id_exist(self.cib.tree, "status-1"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "status-1a"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "status-1aa"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "status-1ab"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "status-1b"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "status-1ba"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "status-1bb"))

    def test_ignore_acl_target(self):
        self.cib.append_to_first_tag_name(
            "configuration",
            """
            <acls>
                <acl_target id="target1"/>
            </acls>
            """,
        )
        self.assertFalse(lib.does_id_exist(self.cib.tree, "target1"))

    def test_ignore_acl_role_references(self):
        self.cib.append_to_first_tag_name(
            "configuration",
            """
            <acls>
                <acl_target id="target1">
                    <role id="role1"/>
                    <role id="role2"/>
                </acl_target>
            </acls>
            """,
        )
        self.assertFalse(lib.does_id_exist(self.cib.tree, "role1"))
        self.assertFalse(lib.does_id_exist(self.cib.tree, "role2"))

    def test_ignore_sections_directly_under_cib(self):
        # this is side effect of current implementation but is not problem since
        # id attribute is not allowed for elements directly under cib
        tree = etree.fromstring('<cib><direct id="a"/></cib>')
        self.assertFalse(lib.does_id_exist(tree, "a"))

    def test_find_id_when_cib_is_not_root_element(self):
        # for example we have only part of xml
        tree = etree.fromstring('<root><direct id="a"/></root>')
        self.assertTrue(lib.does_id_exist(tree, "a"))

    def test_find_remote_node_pacemaker_internal_id(self):
        tree = etree.fromstring(
            """
            <cib>
                <configuration>
                    <resources>
                        <primitive id="b">
                            <meta_attributes>
                                <nvpair name="remote-node" value="a"/>
                            </meta_attributes>
                        </primitive>
                    </resources>
                </configuration>
            </cib>
        """
        )
        self.assertTrue(lib.does_id_exist(tree, "a"))


class FindUniqueIdTest(CibToolsTest):
    def test_already_unique(self):
        self.fixture_add_primitive_with_id("myId")
        self.assertEqual("other", lib.find_unique_id(self.cib.tree, "other"))

    def test_add_suffix(self):
        self.fixture_add_primitive_with_id("myId")
        self.assertEqual("myId-1", lib.find_unique_id(self.cib.tree, "myId"))

        self.fixture_add_primitive_with_id("myId-1")
        self.assertEqual("myId-2", lib.find_unique_id(self.cib.tree, "myId"))

    def test_suffix_not_needed(self):
        self.fixture_add_primitive_with_id("myId-1")
        self.assertEqual("myId", lib.find_unique_id(self.cib.tree, "myId"))

    def test_add_first_available_suffix(self):
        self.fixture_add_primitive_with_id("myId")
        self.fixture_add_primitive_with_id("myId-1")
        self.fixture_add_primitive_with_id("myId-3")
        self.assertEqual("myId-2", lib.find_unique_id(self.cib.tree, "myId"))

    def test_reserved_ids(self):
        self.fixture_add_primitive_with_id("myId-1")
        self.assertEqual(
            "myId-3",
            lib.find_unique_id(self.cib.tree, "myId", ["myId", "myId-2"]),
        )


class CreateSubelementId(TestCase):
    def test_create_plain_id_when_no_conflicting_id_there(self):
        context = etree.fromstring('<cib><a id="b"/></cib>')
        self.assertEqual(
            "b-name",
            lib.create_subelement_id(
                context.find(".//a"), "name", lib.IdProvider(context)
            ),
        )

    def test_create_decorated_id_when_conflicting_id_there(self):
        context = etree.fromstring('<cib><a id="b"><c id="b-name"/></a></cib>')
        self.assertEqual(
            "b-name-1",
            lib.create_subelement_id(
                context.find(".//a"), "name", lib.IdProvider(context)
            ),
        )

    def test_parent_has_no_id(self):
        context = etree.fromstring("<cib><a/></cib>")
        self.assertEqual(
            "a-name",
            lib.create_subelement_id(
                context.find(".//a"), "name", lib.IdProvider(context)
            ),
        )


class GetConfigurationTest(CibToolsTest):
    def test_success_if_exists(self):
        self.assertEqual(
            "configuration", lib.get_configuration(self.cib.tree).tag
        )

    def test_raise_if_missing(self):
        for conf in self.cib.tree.findall(".//configuration"):
            conf.getparent().remove(conf)
        assert_raise_library_error(
            lambda: lib.get_configuration(self.cib.tree),
            (
                severities.ERROR,
                report_codes.CIB_CANNOT_FIND_MANDATORY_SECTION,
                {
                    "section": "configuration",
                },
            ),
        )


class GetConstraintsTest(CibToolsTest):
    def test_success_if_exists(self):
        self.assertEqual("constraints", lib.get_constraints(self.cib.tree).tag)

    def test_raise_if_missing(self):
        for section in self.cib.tree.findall(".//configuration/constraints"):
            section.getparent().remove(section)
        assert_raise_library_error(
            lambda: lib.get_constraints(self.cib.tree),
            (
                severities.ERROR,
                report_codes.CIB_CANNOT_FIND_MANDATORY_SECTION,
                {
                    "section": "configuration/constraints",
                },
            ),
        )


class GetCrmConfig(CibToolsTest):
    def test_success_if_exists(self):
        self.assertEqual("crm_config", lib.get_crm_config(self.cib.tree).tag)

    def test_raise_if_missing(self):
        for section in self.cib.tree.findall(".//configuration/crm_config"):
            section.getparent().remove(section)
        assert_raise_library_error(
            lambda: lib.get_crm_config(self.cib.tree),
            fixture.error(
                report_codes.CIB_CANNOT_FIND_MANDATORY_SECTION,
                section="configuration/crm_config",
            ),
        )


class GetResourcesTest(CibToolsTest):
    def test_success_if_exists(self):
        self.assertEqual("resources", lib.get_resources(self.cib.tree).tag)

    def test_raise_if_missing(self):
        for section in self.cib.tree.findall(".//configuration/resources"):
            section.getparent().remove(section)
        assert_raise_library_error(
            lambda: lib.get_resources(self.cib.tree),
            (
                severities.ERROR,
                report_codes.CIB_CANNOT_FIND_MANDATORY_SECTION,
                {
                    "section": "configuration/resources",
                },
            ),
        )


class GetNodes(CibToolsTest):
    def test_success_if_exists(self):
        self.assertEqual("nodes", lib.get_nodes(self.cib.tree).tag)

    def test_raise_if_missing(self):
        for section in self.cib.tree.findall(".//configuration/nodes"):
            section.getparent().remove(section)
        assert_raise_library_error(
            lambda: lib.get_nodes(self.cib.tree),
            (
                severities.ERROR,
                report_codes.CIB_CANNOT_FIND_MANDATORY_SECTION,
                {
                    "section": "configuration/nodes",
                },
                None,
            ),
        )


class GetAclsTest(CibToolsTest):
    def test_success_if_exists(self):
        self.cib.append_to_first_tag_name(
            "configuration", '<acls><acl_role id="test_role" /></acls>'
        )
        self.assertEqual("test_role", lib.get_acls(self.cib.tree)[0].get("id"))

    def test_success_if_missing(self):
        acls = lib.get_acls(self.cib.tree)
        self.assertEqual("acls", acls.tag)
        self.assertEqual("configuration", acls.getparent().tag)


class GetFencingTopology(CibToolsTest):
    def test_success_if_exists(self):
        self.cib.append_to_first_tag_name(
            "configuration", "<fencing-topology />"
        )
        self.assertEqual(
            "fencing-topology", lib.get_fencing_topology(self.cib.tree).tag
        )

    def test_success_if_missing(self):
        ft_el = lib.get_fencing_topology(self.cib.tree)
        self.assertEqual("fencing-topology", ft_el.tag)
        self.assertEqual("configuration", ft_el.getparent().tag)


class GetTagsTest(CibToolsTest):
    def test_success_if_exists(self):
        self.cib.append_to_first_tag_name("configuration", "<tags/>")
        self.assertEqual("tags", lib.get_tags(self.cib.tree).tag)

    def test_success_if_missing(self):
        tags_el = lib.get_tags(self.cib.tree)
        self.assertEqual("tags", tags_el.tag)
        self.assertEqual("configuration", tags_el.getparent().tag)


@mock.patch("pcs.lib.cib.tools.does_id_exist")
class ValidateIdDoesNotExistsTest(TestCase):
    def test_success_when_id_does_not_exists(self, does_id_exists):
        does_id_exists.return_value = False
        lib.validate_id_does_not_exist("tree", "some-id")
        does_id_exists.assert_called_once_with("tree", "some-id")

    def test_raises_when_id_exists(self, does_id_exists):
        does_id_exists.return_value = True
        assert_raise_library_error(
            lambda: lib.validate_id_does_not_exist("tree", "some-id"),
            (
                severities.ERROR,
                report_codes.ID_ALREADY_EXISTS,
                {"id": "some-id"},
            ),
        )
        does_id_exists.assert_called_once_with("tree", "some-id")


class GetPacemakerVersionByWhichCibWasValidatedTest(TestCase):
    def test_missing_attribute(self):
        assert_raise_library_error(
            lambda: lib.get_pacemaker_version_by_which_cib_was_validated(
                etree.XML("<cib/>")
            ),
            (
                severities.ERROR,
                report_codes.CIB_LOAD_ERROR_BAD_FORMAT,
                {
                    "reason": "the attribute 'validate-with' of the element"
                    " 'cib' is missing"
                },
            ),
        )

    def test_invalid_version(self):
        assert_raise_library_error(
            lambda: lib.get_pacemaker_version_by_which_cib_was_validated(
                etree.XML('<cib validate-with="something-1.2.3"/>')
            ),
            (
                severities.ERROR,
                report_codes.CIB_LOAD_ERROR_BAD_FORMAT,
                {
                    "reason": "the attribute 'validate-with' of the element"
                    " 'cib' has an invalid value: 'something-1.2.3'"
                },
            ),
        )

    def test_invalid_version_at_end(self):
        assert_raise_library_error(
            lambda: lib.get_pacemaker_version_by_which_cib_was_validated(
                etree.XML('<cib validate-with="pacemaker-1.2.3x"/>')
            ),
            (
                severities.ERROR,
                report_codes.CIB_LOAD_ERROR_BAD_FORMAT,
                {
                    "reason": "the attribute 'validate-with' of the element"
                    " 'cib' has an invalid value: 'pacemaker-1.2.3x'"
                },
            ),
        )

    def test_no_revision(self):
        self.assertEqual(
            Version(1, 2),
            lib.get_pacemaker_version_by_which_cib_was_validated(
                etree.XML('<cib validate-with="pacemaker-1.2"/>')
            ),
        )

    def test_with_revision(self):
        self.assertEqual(
            Version(1, 2, 3),
            lib.get_pacemaker_version_by_which_cib_was_validated(
                etree.XML('<cib validate-with="pacemaker-1.2.3"/>')
            ),
        )


class GetCibCrmFeatureSet(TestCase):
    def test_success(self):
        self.assertEqual(
            Version(3, 0, 9),
            lib.get_cib_crm_feature_set(
                etree.XML('<cib crm_feature_set="3.0.9" />')
            ),
        )

    def test_success_no_revision(self):
        self.assertEqual(
            Version(3, 1),
            lib.get_cib_crm_feature_set(
                etree.XML('<cib crm_feature_set="3.1" />')
            ),
        )

    def test_missing_attribute(self):
        assert_raise_library_error(
            lambda: lib.get_cib_crm_feature_set(etree.XML("<cib />")),
            fixture.error(
                report_codes.CIB_LOAD_ERROR_BAD_FORMAT,
                reason=(
                    "the attribute 'crm_feature_set' of the element 'cib' is "
                    "missing"
                ),
            ),
        )

    def test_invalid_version(self):
        assert_raise_library_error(
            lambda: lib.get_cib_crm_feature_set(
                etree.XML('<cib crm_feature_set="3" />')
            ),
            fixture.error(
                report_codes.CIB_LOAD_ERROR_BAD_FORMAT,
                reason=(
                    "the attribute 'crm_feature_set' of the element 'cib' has "
                    "an invalid value: '3'"
                ),
            ),
        )

    def test_invalid_version_at_end(self):
        assert_raise_library_error(
            lambda: lib.get_cib_crm_feature_set(
                etree.XML('<cib crm_feature_set="3.0.9x" />')
            ),
            fixture.error(
                report_codes.CIB_LOAD_ERROR_BAD_FORMAT,
                reason=(
                    "the attribute 'crm_feature_set' of the element 'cib' has "
                    "an invalid value: '3.0.9x'"
                ),
            ),
        )


find_group = partial(lib.find_element_by_tag_and_id, "group")


class FindTagWithId(TestCase):
    def test_returns_element_when_exists(self):
        tree = etree.fromstring(
            '<cib><resources><group id="a"/></resources></cib>'
        )
        element = find_group(tree.find(".//resources"), "a")
        self.assertEqual("group", element.tag)
        self.assertEqual("a", element.attrib["id"])

    def test_returns_element_when_exists_one_of_tags(self):
        tree = etree.fromstring(
            """
            <cib>
                <resources>
                    <group id="a"/>
                    <primitive id="b"/>
                </resources>
            </cib>
        """
        )
        element = lib.find_element_by_tag_and_id(
            ["group", "primitive"], tree.find(".//resources"), "a"
        )
        self.assertEqual("group", element.tag)
        self.assertEqual("a", element.attrib["id"])

    def test_raises_when_is_under_another_tag(self):
        tree = etree.fromstring(
            '<cib><resources><primitive id="a"/></resources></cib>'
        )

        assert_raise_library_error(
            lambda: find_group(tree.find(".//resources"), "a"),
            (
                severities.ERROR,
                report_codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                {
                    "id": "a",
                    "expected_types": ["group"],
                    "current_type": "primitive",
                },
            ),
        )

    def test_raises_when_is_under_another_context(self):
        tree = etree.fromstring(
            """
            <cib>
                <resources>
                    <group id="g1"><primitive id="a"/></group>
                    <group id="g2"><primitive id="b"/></group>
                </resources>
            </cib>
        """
        )
        assert_raise_library_error(
            lambda: lib.find_element_by_tag_and_id(
                "primitive", tree.find('.//resources/group[@id="g2"]'), "a"
            ),
            (
                severities.ERROR,
                report_codes.OBJECT_WITH_ID_IN_UNEXPECTED_CONTEXT,
                {
                    "object_type": "primitive",
                    "object_id": "a",
                    "expected_context_type": "group",
                    "expected_context_id": "g2",
                },
            ),
        )

    def test_raises_when_id_does_not_exists(self):
        tree = etree.fromstring("<cib><resources/></cib>")
        assert_raise_library_error(
            lambda: find_group(tree.find(".//resources"), "a"),
            (
                severities.ERROR,
                report_codes.ID_NOT_FOUND,
                {
                    "id": "a",
                    "expected_types": ["group"],
                    "context_type": "resources",
                    "context_id": "",
                },
                None,
            ),
        )
        assert_raise_library_error(
            lambda: find_group(
                tree.find(".//resources"), "a", id_types=["resource group"]
            ),
            (
                severities.ERROR,
                report_codes.ID_NOT_FOUND,
                {
                    "id": "a",
                    "expected_types": ["resource group"],
                    "context_type": "resources",
                    "context_id": "",
                },
                None,
            ),
        )

    def test_returns_none_if_id_do_not_exists(self):
        tree = etree.fromstring("<cib><resources/></cib>")
        self.assertIsNone(
            find_group(tree.find(".//resources"), "a", none_if_id_unused=True)
        )


class ElementSearcher(TestCase):
    def assert_get_errors_raises(self, searcher):
        with self.assertRaises(AssertionError) as cm:
            searcher.get_errors()
        self.assertEqual(
            str(cm.exception),
            "Improper usage: cannot report errors when there are none",
        )

    def test_element_exists(self):
        tree = etree.fromstring(
            '<cib><resources><group id="a"/></resources></cib>'
        )
        searcher = lib.ElementSearcher("group", "a", tree.find(".//resources"))
        self.assertTrue(searcher.element_found())
        self.assertEqual("a", searcher.get_element().attrib["id"])
        self.assert_get_errors_raises(searcher)

    def test_element_exists_multiple_tags(self):
        tree = etree.fromstring(
            """
            <cib>
                <resources>
                    <group id="a"/>
                    <primitive id="b"/>
                </resources>
            </cib>
        """
        )
        searcher = lib.ElementSearcher(
            ["primitive", "group"], "a", tree.find(".//resources")
        )
        self.assertTrue(searcher.element_found())
        self.assertEqual("a", searcher.get_element().attrib["id"])
        self.assert_get_errors_raises(searcher)

    def test_id_found_for_another_tag(self):
        tree = etree.fromstring(
            '<cib><resources><primitive id="a"/></resources></cib>'
        )
        searcher = lib.ElementSearcher("group", "a", tree.find(".//resources"))
        self.assertFalse(searcher.element_found())
        self.assertIsNone(searcher.get_element())
        assert_report_item_list_equal(
            searcher.get_errors(),
            [
                fixture.error(
                    report_codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="a",
                    expected_types=["group"],
                    current_type="primitive",
                ),
            ],
        )

    def test_element_exists_in_another_context(self):
        tree = etree.fromstring(
            """
            <cib>
                <resources>
                    <group id="g1"><primitive id="a"/></group>
                    <group id="g2"><primitive id="b"/></group>
                </resources>
            </cib>
        """
        )
        searcher = lib.ElementSearcher(
            "primitive",
            "a",
            tree.find('.//resources/group[@id="g2"]'),
        )
        self.assertFalse(searcher.element_found())
        self.assertIsNone(searcher.get_element())
        assert_report_item_list_equal(
            searcher.get_errors(),
            [
                fixture.error(
                    report_codes.OBJECT_WITH_ID_IN_UNEXPECTED_CONTEXT,
                    object_id="a",
                    object_type="primitive",
                    expected_context_type="group",
                    expected_context_id="g2",
                ),
            ],
        )

    def assert_id_does_not_exists(self, element_type_desc, expected_types):
        tree = etree.fromstring("<cib><resources/></cib>")
        searcher = lib.ElementSearcher(
            "group",
            "a",
            tree.find(".//resources"),
            element_type_desc=element_type_desc,
        )
        self.assertFalse(searcher.element_found())
        self.assertIsNone(searcher.get_element())
        assert_report_item_list_equal(
            searcher.get_errors(),
            [
                fixture.error(
                    report_codes.ID_NOT_FOUND,
                    id="a",
                    expected_types=expected_types,
                    context_type="resources",
                    context_id="",
                ),
            ],
        )

    def test_id_does_not_exists(self):
        self.assert_id_does_not_exists(None, ["group"])

    def test_id_does_not_exists_custom_description_string(self):
        self.assert_id_does_not_exists("resource group", ["resource group"])

    def test_id_does_not_exists_custom_description_list(self):
        self.assert_id_does_not_exists(
            ["resource", "group"], ["group", "resource"]
        )

    def test_book_available_valid_id(self):
        tree = etree.fromstring(
            '<cib><resources><group id="b"/></resources></cib>'
        )
        id_provider = lib.IdProvider(tree)
        searcher = lib.ElementSearcher("group", "a", tree.find(".//resources"))
        self.assertFalse(searcher.element_found())
        self.assertTrue(searcher.validate_book_id(id_provider))
        self.assert_get_errors_raises(searcher)

    def test_book_valid_id_used_in_cib(self):
        tree = etree.fromstring(
            '<cib><resources><group id="a"/></resources></cib>'
        )
        id_provider = lib.IdProvider(tree)
        searcher = lib.ElementSearcher("group", "a", tree.find(".//resources"))
        self.assertTrue(searcher.element_found())
        self.assertFalse(searcher.validate_book_id(id_provider))
        # does not report book errors  - the element has been found,
        # validate_book_id should have not been called
        self.assert_get_errors_raises(searcher)

    def test_book_valid_id_used_in_id_provider(self):
        tree = etree.fromstring(
            '<cib><resources><group id="b"/></resources></cib>'
        )
        id_provider = lib.IdProvider(tree)
        self.assertEqual([], id_provider.book_ids("a"))
        searcher = lib.ElementSearcher("group", "a", tree.find(".//resources"))
        self.assertFalse(searcher.element_found())
        self.assertFalse(searcher.validate_book_id(id_provider))
        assert_report_item_list_equal(
            searcher.get_errors(),
            [
                fixture.error(
                    report_codes.ID_ALREADY_EXISTS,
                    id="a",
                ),
            ],
        )

    def test_book_not_valid_id(self):
        tree = etree.fromstring(
            '<cib><resources><group id="b"/></resources></cib>'
        )
        id_provider = lib.IdProvider(tree)
        searcher = lib.ElementSearcher("group", "1a", tree.find(".//resources"))
        self.assertFalse(searcher.element_found())
        self.assertFalse(searcher.validate_book_id(id_provider, "group name"))
        assert_report_item_list_equal(
            searcher.get_errors(),
            [
                fixture.error(
                    report_codes.INVALID_ID_BAD_CHAR,
                    id="1a",
                    id_description="group name",
                    is_first_char=True,
                    invalid_character="1",
                ),
            ],
        )


class GetElementById(TestCase):
    def test_one_match(self):
        found_element = lib.get_element_by_id(cib_element_lookup, "R1")
        self.assertEqual(
            found_element,
            cib_element_lookup.find('.//resources/primitive[@id="R1"]'),
        )

    def test_one_match_ignore_reference(self):
        found_element = lib.get_element_by_id(cib_element_lookup, "R3")
        self.assertEqual(
            found_element,
            cib_element_lookup.find('.//resources/primitive[@id="R3"]'),
        )

    def test_duplicate_ids(self):
        with self.assertRaises(AssertionError):
            lib.get_element_by_id(cib_element_lookup, "RX2")

    def test_tag_id_same_as_resource_id(self):
        with self.assertRaises(AssertionError):
            lib.get_element_by_id(cib_element_lookup, "T")

    def test_id_not_in_cib(self):
        with self.assertRaises(lib.ElementNotFound):
            lib.get_element_by_id(cib_element_lookup, "X")

    def test_no_match_in_status(self):
        with self.assertRaises(lib.ElementNotFound):
            lib.get_element_by_id(cib_element_lookup, "R2")

    def test_no_match_in_obj_ref(self):
        with self.assertRaises(lib.ElementNotFound):
            lib.get_element_by_id(cib_element_lookup, "RX1")


class GetElementsById(TestCase):
    def assert_result(self, found_id_list, unmatched_id_list):
        found_elements, unmatched_ids = lib.get_elements_by_ids(
            cib_element_lookup, found_id_list + unmatched_id_list
        )
        self.assertListEqual(
            found_elements,
            [
                cib_element_lookup.find(
                    f'.//resources/primitive[@id="{found_id}"]'
                )
                for found_id in found_id_list
            ],
        )
        self.assertListEqual(unmatched_ids, unmatched_id_list)

    def test_one_match(self):
        self.assert_result(["R1"], [])

    def test_one_match_ignore_reference(self):
        self.assert_result(["R3"], [])

    def test_duplicate_ids(self):
        with self.assertRaises(AssertionError):
            lib.get_elements_by_ids(cib_element_lookup, ["RX2"])

    def test_tag_id_same_as_resource_id(self):
        with self.assertRaises(AssertionError):
            lib.get_elements_by_ids(cib_element_lookup, ["T"])

    def test_id_not_in_cib(self):
        self.assert_result(["R1"], ["X1", "X2"])

    def test_no_match_in_status(self):
        self.assert_result([], ["R2"])

    def test_no_match_in_obj_ref(self):
        self.assert_result([], ["RX1"])



import re
from typing import (
    Optional,
    Union,
)

_INTEGER_RE = re.compile(r"^[+-]?[0-9]+$")


def is_integer(
    value: Union[str, int, float],
    at_least: Optional[int] = None,
    at_most: Optional[int] = None,
) -> bool:
    """
    Check if the specified value is an integer, optionally check a range

    value -- value to check
    at_least -- minimal allowed value
    at_most -- maximal allowed value
    """
    try:
        if value is None or isinstance(value, float):
            return False
        if isinstance(value, str) and not _INTEGER_RE.fullmatch(value):
            return False
        value_int = int(value)
        if at_least is not None and value_int < at_least:
            return False
        if at_most is not None and value_int > at_most:
            return False
    except ValueError:
        return False
    return True


def is_port_number(value: str) -> bool:
    """
    Check if the specified value is a TCP or UDP port number

    value -- value to check
    """
    return is_integer(value, 1, 65535)



import sys
import textwrap
from shutil import get_terminal_size
from typing import List

INDENT_STEP = 2
SUBSEQUENT_INDENT_STEP = 4


def _smart_wrap(
    text: str, subsequent_indent: int = SUBSEQUENT_INDENT_STEP
) -> List[str]:
    initial_indent = len(text) - len(text.lstrip(" "))
    return format_wrap_for_terminal(
        text, subsequent_indent=subsequent_indent + initial_indent
    )


def smart_wrap_text(
    lines: List[str], subsequent_indent: int = SUBSEQUENT_INDENT_STEP
) -> List[str]:
    output = []
    for line in lines:
        if not line:
            output.append("")
            continue
        output.extend(_smart_wrap(line, subsequent_indent=subsequent_indent))
    return output


def format_wrap_for_terminal(
    text: str,
    subsequent_indent: int = SUBSEQUENT_INDENT_STEP,
    trim: int = 0,
) -> List[str]:
    """
    Returns text as a list of lines. Length of a line is determined by a
    terminal size if not explicitly specified.

    text -- string to format
    subsequent_indent -- number of spaces all subsequent lines will be indented
        compared to the first one.
    trim -- number which will be substracted from terminal size. Can be used in
        cases lines will be indented later by this number of spaces.
    """
    if any((sys.stdout.isatty(), sys.stderr.isatty())):
        return format_wrap(
            text,
            # minimal line length is 40
            max(get_terminal_size()[0] - trim, 40),
            subsequent_indent=subsequent_indent,
        )
    return [text]


def format_wrap(
    text: str,
    max_length: int,
    subsequent_indent: int = SUBSEQUENT_INDENT_STEP,
) -> List[str]:
    return textwrap.wrap(
        text,
        max_length,
        subsequent_indent=" " * subsequent_indent,
    )



from dataclasses import replace
from typing import (
    List,
    Optional,
    Tuple,
)

from pcs.common.const import PcmkRoleType
from pcs.common.pacemaker import role

from . import const
from .types import (
    ResourceAgentAction,
    ResourceAgentMetadata,
    ResourceAgentParameter,
)


def get_additional_trace_parameters(
    existing_parameters: List[ResourceAgentParameter],
) -> List[ResourceAgentParameter]:
    """
    Return trace parameters which need to be added based on existing parameters

    existing_parameters -- parameters defined by an agent
    """
    trace_ra_found = trace_file_found = False
    for param in existing_parameters:
        param_name = param.name.lower()
        if param_name == "trace_ra":
            trace_ra_found = True
        elif param_name == "trace_file":
            trace_file_found = True
        if trace_file_found and trace_ra_found:
            break

    result = []
    if not trace_ra_found:
        shortdesc = (
            "Set to 1 to turn on resource agent tracing (expect large output)"
        )
        result.append(
            ResourceAgentParameter(
                name="trace_ra",
                shortdesc=shortdesc,
                longdesc=(
                    shortdesc + " The trace output will be saved to "
                    "trace_file, if set, or by default to "
                    "$HA_VARRUN/ra_trace/<type>/<id>.<action>.<timestamp> e.g. "
                    "$HA_VARRUN/ra_trace/oracle/db.start.2012-11-27.08:37:08"
                ),
                type="integer",
                default="0",
                enum_values=None,
                required=False,
                advanced=True,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            )
        )
    if not trace_file_found:
        shortdesc = "Path to a file to store resource agent tracing log"
        result.append(
            ResourceAgentParameter(
                name="trace_file",
                shortdesc=shortdesc,
                longdesc=shortdesc,
                type="string",
                default="",
                enum_values=None,
                required=False,
                advanced=True,
                deprecated=False,
                deprecated_by=[],
                deprecated_desc=None,
                unique_group=None,
                reloadable=False,
            )
        )
    return result


def ocf_unified_to_pcs(
    metadata: ResourceAgentMetadata,
) -> ResourceAgentMetadata:
    """
    Return a cleaned up agent metadata suitable for further pcs processing

    metadata -- parsed OCF agent metadata to be cleaned
    """
    result = metadata
    result = _metadata_action_translate_role(result)
    if metadata.name.is_pcmk_fake_agent:
        result = _metadata_parameter_extract_enum_values_from_desc(result)
        result = _metadata_parameter_remove_select_enum_values_from_desc(result)
        result = _metadata_parameter_deduplicate_desc(result)
        result = _metadata_parameter_join_short_long_desc(result)
        result = _metadata_parameter_extract_advanced_from_desc(result)
    if metadata.name.is_stonith:
        result = _metadata_remove_unwanted_stonith_parameters(result)
        result = _metadata_make_stonith_action_parameter_deprecated(result)
        result = _metadata_make_stonith_port_parameter_not_required(result)
    return result


def _metadata_action_translate_role(
    metadata: ResourceAgentMetadata,
) -> ResourceAgentMetadata:
    return replace(
        metadata,
        actions=[_action_translate_role(action) for action in metadata.actions],
    )


def _action_translate_role(action: ResourceAgentAction) -> ResourceAgentAction:
    if action.role is None:
        return action
    return replace(
        action, role=role.get_value_primary(PcmkRoleType(action.role))
    )


def _metadata_parameter_extract_advanced_from_desc(
    metadata: ResourceAgentMetadata,
) -> ResourceAgentMetadata:
    return replace(
        metadata,
        parameters=[
            _parameter_extract_advanced_from_desc(parameter)
            for parameter in metadata.parameters
        ],
    )


def _parameter_extract_advanced_from_desc(
    parameter: ResourceAgentParameter,
) -> ResourceAgentParameter:
    advanced_str_list = ["Advanced use only: ", "*** Advanced Use Only *** "]
    if parameter.shortdesc:
        for advanced_str in advanced_str_list:
            if parameter.shortdesc.startswith(advanced_str):
                return replace(parameter, advanced=True)
    return parameter


def _metadata_parameter_join_short_long_desc(
    metadata: ResourceAgentMetadata,
) -> ResourceAgentMetadata:
    return replace(
        metadata,
        parameters=[
            _parameter_join_short_long_desc(parameter)
            for parameter in metadata.parameters
        ],
    )


def _parameter_join_short_long_desc(
    parameter: ResourceAgentParameter,
) -> ResourceAgentParameter:
    shortdesc = parameter.shortdesc
    longdesc = parameter.longdesc
    if shortdesc is None or longdesc is None:
        return parameter
    return replace(parameter, longdesc=f"{shortdesc}\n{longdesc}".strip())


def _metadata_remove_unwanted_stonith_parameters(
    metadata: ResourceAgentMetadata,
) -> ResourceAgentMetadata:
    # We don't allow the user to change these options which are only intended
    # to be used interactively on command line.
    return replace(
        metadata,
        parameters=[
            parameter
            for parameter in metadata.parameters
            if parameter.name not in {"help", "version"}
        ],
    )


def _metadata_make_stonith_action_parameter_deprecated(
    metadata: ResourceAgentMetadata,
) -> ResourceAgentMetadata:
    # Action parameter is intended to be used interactively on command line
    # only. However, we still need the user to be able to set it due to
    # backward compatibility reasons. So we just mark it as not required. We
    # also move it to advanced params to indicate that users should not set it
    # in most cases.
    new_parameters = []
    for param in metadata.parameters:
        if param.name == "action":
            new_parameters.append(
                replace(
                    param,
                    required=False,
                    advanced=True,
                    deprecated=True,
                    deprecated_by=(
                        const.STONITH_ACTION_REPLACED_BY + param.deprecated_by
                    ),
                )
            )
        else:
            new_parameters.append(param)
    return replace(metadata, parameters=new_parameters)


def _metadata_make_stonith_port_parameter_not_required(
    metadata: ResourceAgentMetadata,
) -> ResourceAgentMetadata:
    # 'port' parameter is required by a fence agent, but it is filled
    # automatically by pacemaker based on 'pcmk_host_map' or 'pcmk_host_list'
    # parameter (defined in fenced metadata). Therefore, we must mark 'port'
    # and all parameters replacing it as not required.
    port_related_params = set()
    next_iteration_params = set(["port"])
    while next_iteration_params:
        current_params = next_iteration_params
        next_iteration_params = set()
        for param in metadata.parameters:
            if (
                param.name in current_params
                and param.name not in port_related_params
            ):
                port_related_params.add(param.name)
                next_iteration_params.update(param.deprecated_by)

    new_parameters = []
    for param in metadata.parameters:
        if param.name in port_related_params:
            new_parameters.append(replace(param, required=False))
        else:
            new_parameters.append(param)
    return replace(metadata, parameters=new_parameters)


def _metadata_parameter_extract_enum_values_from_desc(
    metadata: ResourceAgentMetadata,
) -> ResourceAgentMetadata:
    return replace(
        metadata,
        parameters=[
            _parameter_extract_enum_values_from_desc(parameter)
            for parameter in metadata.parameters
        ],
    )


def _parameter_extract_enum_values_from_desc(
    parameter: ResourceAgentParameter,
) -> ResourceAgentParameter:
    if parameter.type != "enum":
        return parameter
    enum_values, longdesc = _get_enum_values_and_new_longdesc(parameter)
    if parameter.default is not None and parameter.default not in enum_values:
        enum_values.append(parameter.default)
    parameter = replace(parameter, type="select")
    parameter = replace(parameter, longdesc=longdesc)
    return replace(parameter, enum_values=enum_values)


def _metadata_parameter_remove_select_enum_values_from_desc(
    metadata: ResourceAgentMetadata,
) -> ResourceAgentMetadata:
    return replace(
        metadata,
        parameters=[
            _parameter_remove_select_enum_values_from_desc(parameter)
            for parameter in metadata.parameters
        ],
    )


def _parameter_remove_select_enum_values_from_desc(
    parameter: ResourceAgentParameter,
) -> ResourceAgentParameter:
    if parameter.type != "select":
        return parameter
    _, longdesc = _get_enum_values_and_new_longdesc(parameter)
    return replace(parameter, longdesc=longdesc)


def _metadata_parameter_deduplicate_desc(
    metadata: ResourceAgentMetadata,
) -> ResourceAgentMetadata:
    return replace(
        metadata,
        parameters=[
            _parameter_deduplicate_desc(parameter)
            for parameter in metadata.parameters
        ],
    )


def _parameter_deduplicate_desc(
    parameter: ResourceAgentParameter,
) -> ResourceAgentParameter:
    if parameter.shortdesc == parameter.longdesc:
        return replace(parameter, longdesc=None)
    return parameter


def _get_enum_values_and_new_longdesc(
    parameter: ResourceAgentParameter,
) -> Tuple[List[str], Optional[str]]:
    enum_values = []
    longdesc = parameter.longdesc
    if parameter.longdesc:
        parts = parameter.longdesc.split("  Allowed values: ")
        if len(parts) == 2:
            enum_values = parts[1].split(", ")
            longdesc = parts[0]
    return enum_values, longdesc



from unittest import (
    TestCase,
    mock,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.constraint.parse_args import (
    prepare_resource_sets,
    prepare_set_args,
)


@mock.patch("pcs.cli.common.parse_args.prepare_options")
class PrepareResourceSetsTest(TestCase):
    def test_prepare_resource_sets(self, options):
        opts = [{"id": "1"}, {"id": "2", "sequential": "true"}]
        options.side_effect = opts
        self.assertEqual(
            [
                {"ids": ["resA", "resB"], "options": opts[0]},
                {"ids": ["resC"], "options": opts[1]},
            ],
            prepare_resource_sets(
                [
                    "resA",
                    "resB",
                    "id=resource-set-1",
                    "set",
                    "resC",
                    "id=resource-set-2",
                    "sequential=true",
                ]
            ),
        )

    def test_has_no_responsibility_to_assess_the_content(self, options):
        options.return_value = {}
        self.assertEqual(
            [{"ids": [], "options": {}}], prepare_resource_sets([])
        )


@mock.patch("pcs.cli.common.parse_args.prepare_options")
@mock.patch("pcs.cli.constraint.parse_args.prepare_resource_sets")
class PrepareSetArgvTest(TestCase):
    # pylint: disable=no-self-use, unused-argument
    def test_return_tuple_of_given_resource_set_list_and_options(
        self, res_sets, options
    ):
        res_sets.return_value = [{"ids": "A"}]
        options.return_value = "O"

        self.assertEqual(
            ([{"ids": "A"}], "O"),
            prepare_set_args(["A", "b=c", "setoptions", "d=e"]),
        )

    def test_right_distribute_full_args(self, res_sets, options):
        prepare_set_args(["A", "b=c", "setoptions", "d=e"])
        res_sets.assert_called_once_with(["A", "b=c"])
        options.assert_called_once_with(["d=e"])

    def test_right_distribute_args_without_options(self, res_sets, options):
        prepare_set_args(["A", "b=c"])
        res_sets.assert_called_once_with(["A", "b=c"])
        options.assert_not_called()

    def test_right_distribute_args_with_empty_options(self, res_sets, options):
        prepare_set_args(["A", "b=c", "setoptions"])
        res_sets.assert_called_once_with(["A", "b=c"])
        options.assert_not_called()

    def test_raises_when_no_set_specified(self, res_sets, options):
        self.assertRaises(CmdLineInputError, lambda: prepare_set_args([]))
        res_sets.assert_not_called()

    def test_raises_when_no_resource_in_set(self, res_sets, options):
        res_sets.return_value = [{"ids": [], "options": {"b": "c"}}]
        self.assertRaises(CmdLineInputError, lambda: prepare_set_args(["b=c"]))
        res_sets.assert_called_once_with(["b=c"])

    def test_raises_when_setoption_more_than_once(self, res_sets, options):
        self.assertRaises(
            CmdLineInputError,
            lambda: prepare_set_args(
                ["A", "b=c", "setoptions", "c=d", "setoptions", "e=f"]
            ),
        )



from unittest import TestCase

from lxml import etree

from pcs.common import reports
from pcs.lib.cib.resource import validations

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal


def _resource(cib, id_):
    return cib.find(f".//*[@id='{id_}']")


def _resources(cib, *ids):
    return [_resource(cib, id_) for id_ in ids]


class ValidateMoveResourcesToGroup(TestCase):
    def setUp(self):
        self.cib = etree.fromstring(
            """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="RG2" />
                </group>
                <group id="GX">
                    <primitive id="RGX" />
                </group>
                <primitive id="R1" />
                <primitive id="R2" />
                <primitive id="R3" />
                <clone id="RC1-clone">
                    <primitive id="RC1" />
                </clone>
                <bundle id="RB1-bundle">
                    <primitive id="RB1">
                        <meta_attributes id="RB1-meta_attributes" />
                    </primitive>
                </bundle>
            </resources>
        """
        )

    def _validate(self, group, resources, adjacent=None):
        return validations.validate_move_resources_to_group(
            _resource(self.cib, group),
            _resources(self.cib, *resources),
            _resource(self.cib, adjacent) if adjacent else None,
        )

    def test_no_resources_specified(self):
        assert_report_item_list_equal(
            self._validate("G", []),
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_ITEMS_NOT_SPECIFIED,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_GROUP,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_RESOURCE,
                    container_id="G",
                ),
            ],
        )

    def test_group_is_not_group(self):
        assert_report_item_list_equal(
            self._validate("RB1-meta_attributes", ["R1"]),
            [
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="RB1-meta_attributes",
                    expected_types=["group"],
                    current_type="meta_attributes",
                ),
            ],
        )

    def test_resources_are_not_primitives(self):
        assert_report_item_list_equal(
            self._validate("G", ["RC1-clone", "R1", "RB1-bundle"]),
            [
                fixture.error(
                    reports.codes.CANNOT_GROUP_RESOURCE_WRONG_TYPE,
                    resource_id="RC1-clone",
                    resource_type="clone",
                    parent_id=None,
                    parent_type=None,
                ),
                fixture.error(
                    reports.codes.CANNOT_GROUP_RESOURCE_WRONG_TYPE,
                    resource_id="RB1-bundle",
                    resource_type="bundle",
                    parent_id=None,
                    parent_type=None,
                ),
            ],
        )

    def test_resources_are_in_clones_etc(self):
        assert_report_item_list_equal(
            self._validate("G", ["RC1", "R1", "RB1"]),
            [
                fixture.error(
                    reports.codes.CANNOT_GROUP_RESOURCE_WRONG_TYPE,
                    resource_id="RC1",
                    resource_type="primitive",
                    parent_id="RC1-clone",
                    parent_type="clone",
                ),
                fixture.error(
                    reports.codes.CANNOT_GROUP_RESOURCE_WRONG_TYPE,
                    resource_id="RB1",
                    resource_type="primitive",
                    parent_id="RB1-bundle",
                    parent_type="bundle",
                ),
            ],
        )

    def test_resources_already_in_the_group(self):
        assert_report_item_list_equal(
            self._validate("G", ["RG2", "R1", "RG1"]),
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_CANNOT_ADD_ITEMS_ALREADY_IN_THE_CONTAINER,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_GROUP,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_RESOURCE,
                    container_id="G",
                    item_list=["RG1", "RG2"],
                ),
            ],
        )

    def test_allow_moving_resources_in_a_group_if_adjacent(self):
        assert_report_item_list_equal(
            self._validate("G", ["RG2", "R1"], "RG1"), []
        )

    def test_adjacent_resource_not_in_the_group(self):
        # pylint: disable=line-too-long
        assert_report_item_list_equal(
            self._validate("G", ["R1"], "R2"),
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_ADJACENT_ITEM_NOT_IN_THE_CONTAINER,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_GROUP,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_RESOURCE,
                    container_id="G",
                    adjacent_item_id="R2",
                ),
            ],
        )

    def test_adjacent_resource_in_another_group(self):
        # pylint: disable=line-too-long
        assert_report_item_list_equal(
            self._validate("G", ["R1"], "RGX"),
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_ADJACENT_ITEM_NOT_IN_THE_CONTAINER,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_GROUP,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_RESOURCE,
                    container_id="G",
                    adjacent_item_id="RGX",
                ),
            ],
        )

    def test_adjacent_resource_to_be_grouped(self):
        assert_report_item_list_equal(
            self._validate("G", ["RG1"], "RG1"),
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_CANNOT_PUT_ITEM_NEXT_TO_ITSELF,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_GROUP,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_RESOURCE,
                    container_id="G",
                    adjacent_item_id="RG1",
                ),
            ],
        )

    def test_resources_specified_twice(self):
        assert_report_item_list_equal(
            self._validate("G", ["R3", "R2", "R1", "R2", "R1"]),
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_ITEMS_DUPLICATION,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_GROUP,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_RESOURCE,
                    container_id="G",
                    duplicate_items_list=["R1", "R2"],
                ),
            ],
        )

    def test_resources_are_not_resources(self):
        assert_report_item_list_equal(
            self._validate("G", ["RB1-meta_attributes"]),
            [
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="RB1-meta_attributes",
                    expected_types=["primitive"],
                    current_type="meta_attributes",
                ),
            ],
        )

    def test_adjacent_same_as_moved(self):
        assert_report_item_list_equal(
            self._validate("G", ["RG1", "RG2"], "RG1"),
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_CANNOT_PUT_ITEM_NEXT_TO_ITSELF,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_GROUP,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_RESOURCE,
                    container_id="G",
                    adjacent_item_id="RG1",
                ),
            ],
        )

    def test_adjacent_same_as_moved_new_group(self):
        empty_group_element = etree.fromstring('<group id="G-new" />')
        assert_report_item_list_equal(
            validations.validate_move_resources_to_group(
                empty_group_element,
                _resources(self.cib, "RG1"),
                _resource(self.cib, "RG1"),
            ),
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_CANNOT_PUT_ITEM_NEXT_TO_ITSELF,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_GROUP,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_RESOURCE,
                    container_id="G-new",
                    adjacent_item_id="RG1",
                ),
                fixture.error(
                    reports.codes.ADD_REMOVE_ADJACENT_ITEM_NOT_IN_THE_CONTAINER,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_GROUP,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_RESOURCE,
                    container_id="G-new",
                    adjacent_item_id="RG1",
                ),
            ],
        )

    def test_new_group_not_valid_id(self):
        # TODO: This can no longer be tested here
        """
        assert_report_item_list_equal(
            self._validate("1Gr:oup", ["R1"]),
            [
                fixture.error(
                    reports.codes.INVALID_ID_BAD_CHAR,
                    id="1Gr:oup",
                    id_description="group name",
                    is_first_char=True,
                    invalid_character="1",
                ),
                fixture.error(
                    reports.codes.INVALID_ID_BAD_CHAR,
                    id="1Gr:oup",
                    id_description="group name",
                    is_first_char=False,
                    invalid_character=":",
                ),
            ],
        )
        """

    def test_adjacent_resource_new_group(self):
        empty_group_element = etree.fromstring('<group id="G-new" />')
        assert_report_item_list_equal(
            validations.validate_move_resources_to_group(
                empty_group_element,
                _resources(self.cib, "RG1"),
                _resource(self.cib, "RG2"),
            ),
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_ADJACENT_ITEM_NOT_IN_THE_CONTAINER,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_GROUP,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_RESOURCE,
                    container_id="G-new",
                    adjacent_item_id="RG2",
                ),
            ],
        )

    def test_adjacent_resource_doesnt_exist(self):
        # TODO: This can no longer be tested here
        """

        assert_report_item_list_equal(
            self._validate("G", "RX"),
            [
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id="RX",
                    expected_types=["primitive"],
                    context_type="resources",
                    context_id="",
                ),
            ],
        )
        """


class ValidateMoveBanClearMixin:
    # pylint: disable=too-many-public-methods,line-too-long
    @staticmethod
    def _fixture_clone(promotable=False):
        return etree.fromstring(
            f"""
            <clone id="R-clone">
                <primitive id="R" />
                <meta_attributes>
                    <nvpair name="promotable" value="{'true' if promotable else 'false'}" />
                </meta_attributes>
            </clone>
            """
        )

    @staticmethod
    def _fixture_group_clone(promotable=False):
        return etree.fromstring(
            f"""
            <clone id="G-clone">
                <group id="G">
                    <primitive id="R" />
                </group>
                <meta_attributes>
                    <nvpair name="promotable" value="{'true' if promotable else 'false'}" />
                </meta_attributes>
            </clone>
            """
        )

    @staticmethod
    def _fixture_master():
        return etree.fromstring(
            """
            <master id="R-master">
                <primitive id="R" />
            </master>
            """
        )

    @staticmethod
    def _fixture_group_master():
        return etree.fromstring(
            """
            <master id="G-master">
                <group id="G">
                    <primitive id="R" />
                </group>
            </master>
            """
        )

    def test_master_true_promotable_clone(self):
        element = self._fixture_clone(True)
        assert_report_item_list_equal(self.validate(element, True), [])

    def test_master_false_promotable_clone(self):
        element = self._fixture_clone(True)
        assert_report_item_list_equal(self.validate(element, False), [])

    def test_master_true_clone(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(
            self.validate(element, True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="R-clone",
                    promotable_id=None,
                ),
            ],
        )

    def test_master_false_clone(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(self.validate(element, False), [])

    def test_master_true_master(self):
        element = self._fixture_master()
        assert_report_item_list_equal(self.validate(element, True), [])

    def test_master_false_master(self):
        element = self._fixture_master()
        assert_report_item_list_equal(self.validate(element, False), [])

    def test_master_true_promotable_clone_resource(self):
        element = self._fixture_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="R",
                    promotable_id="R-clone",
                ),
            ],
        )

    def test_master_false_promotable_clone_resource(self):
        element = self._fixture_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), False), []
        )

    def test_master_true_promotable_clone_group(self):
        element = self._fixture_group_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./group"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="G",
                    promotable_id="G-clone",
                ),
            ],
        )

    def test_master_false_promotable_clone_group(self):
        element = self._fixture_group_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./group"), False), []
        )

    def test_master_true_promotable_clone_group_resource(self):
        element = self._fixture_group_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="R",
                    promotable_id="G-clone",
                ),
            ],
        )

    def test_master_false_promotable_clone_group_resource(self):
        element = self._fixture_group_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), False), []
        )

    def test_master_true_clone_resource(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="R",
                    promotable_id=None,
                ),
            ],
        )

    def test_master_false_clone_resource(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), False), []
        )

    def test_master_true_clone_group(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="G",
                    promotable_id=None,
                ),
            ],
        )

    def test_master_false_clone_group(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group"), False), []
        )

    def test_master_true_clone_group_resource(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="R",
                    promotable_id=None,
                ),
            ],
        )

    def test_master_false_clone_group_resource(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), False), []
        )

    def test_master_true_master_resource(self):
        element = self._fixture_master()
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="R",
                    promotable_id="R-master",
                ),
            ],
        )

    def test_master_true_master_group(self):
        element = self._fixture_group_master()
        assert_report_item_list_equal(
            self.validate(element.find("./group"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="G",
                    promotable_id="G-master",
                ),
            ],
        )

    def test_master_true_master_group_resource(self):
        element = self._fixture_group_master()
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), True),
            [
                fixture.error(
                    self.report_code_bad_master,
                    resource_id="R",
                    promotable_id="G-master",
                ),
            ],
        )


class ValidateMove(ValidateMoveBanClearMixin, TestCase):
    validate = staticmethod(validations.validate_move)
    report_code_bad_master = (
        reports.codes.CANNOT_MOVE_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE
    )

    @staticmethod
    def _fixture_bundle():
        return etree.fromstring(
            """
            <bundle id="R-bundle">
                <primitive id="R" />
            </bundle>
            """
        )

    def test_master_false_promotable_clone(self):
        element = self._fixture_clone(True)
        assert_report_item_list_equal(
            self.validate(element, False),
            [],
        )

    def test_master_true_promotable_clone(self):
        element = self._fixture_clone(True)
        assert_report_item_list_equal(
            self.validate(element, True),
            [],
        )

    def test_master_true_clone(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(
            self.validate(element, True),
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="R-clone",
                ),
            ],
        )

    def test_master_false_clone(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(
            self.validate(element, False),
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="R-clone",
                ),
            ],
        )

    def test_master_false_master(self):
        element = self._fixture_master()
        assert_report_item_list_equal(
            self.validate(element, False),
            [],
        )

    def test_master_true_master(self):
        element = self._fixture_master()
        assert_report_item_list_equal(
            self.validate(element, True),
            [],
        )

    def test_master_false_promotable_clone_resource(self):
        element = self._fixture_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), False),
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_PROMOTABLE_INNER,
                    resource_id="R",
                    promotable_id="R-clone",
                ),
            ],
        )

    def test_master_false_promotable_clone_group(self):
        element = self._fixture_group_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./group"), False),
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_PROMOTABLE_INNER,
                    resource_id="G",
                    promotable_id="G-clone",
                ),
            ],
        )

    def test_master_false_promotable_clone_group_resource(self):
        element = self._fixture_group_clone(True)
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), False),
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_PROMOTABLE_INNER,
                    resource_id="R",
                    promotable_id="G-clone",
                ),
            ],
        )

    def test_master_true_clone_resource(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), True),
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="R",
                ),
            ],
        )

    def test_master_false_clone_resource(self):
        element = self._fixture_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), False),
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="R",
                ),
            ],
        )

    def test_master_true_clone_group(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group"), True),
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="G",
                ),
            ],
        )

    def test_master_false_clone_group(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group"), False),
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="G",
                ),
            ],
        )

    def test_master_true_clone_group_resource(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), True),
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="R",
                ),
            ],
        )

    def test_master_false_clone_group_resource(self):
        element = self._fixture_group_clone(False)
        assert_report_item_list_equal(
            self.validate(element.find("./group/primitive"), False),
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_CLONE,
                    resource_id="R",
                ),
            ],
        )

    def test_bundle(self):
        element = self._fixture_bundle()
        assert_report_item_list_equal(
            self.validate(element, False),
            [
                fixture.error(
                    reports.codes.CANNOT_MOVE_RESOURCE_BUNDLE,
                    resource_id="R-bundle",
                ),
            ],
        )

    def test_bundle_resource(self):
        element = self._fixture_bundle()
        assert_report_item_list_equal(
            self.validate(element.find("./primitive"), False), []
        )


class ValidateBan(ValidateMoveBanClearMixin, TestCase):
    validate = staticmethod(validations.validate_ban)
    report_code_bad_master = (
        reports.codes.CANNOT_BAN_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE
    )


class ValidateUnmoveUnban(ValidateMoveBanClearMixin, TestCase):
    validate = staticmethod(validations.validate_unmove_unban)
    report_code_bad_master = (
        reports.codes.CANNOT_UNMOVE_UNBAN_RESOURCE_MASTER_RESOURCE_NOT_PROMOTABLE
    )



from pcs.common.resource_agent.dto import (
    ListResourceAgentNameDto,
    ResourceAgentActionDto,
    ResourceAgentMetadataDto,
    ResourceAgentNameDto,
    ResourceAgentParameterDto,
)

from .error import (
    AgentNameGuessFoundMoreThanOne,
    AgentNameGuessFoundNone,
    InvalidResourceAgentName,
    ResourceAgentError,
    UnableToGetAgentMetadata,
    UnsupportedOcfVersion,
    resource_agent_error_to_report_item,
)
from .facade import (
    ResourceAgentFacade,
    ResourceAgentFacadeFactory,
)
from .list import (
    find_one_resource_agent_by_type,
    list_resource_agents,
    list_resource_agents_ocf_providers,
    list_resource_agents_standards,
    list_resource_agents_standards_and_providers,
)
from .name import split_resource_agent_name
from .types import (
    ResourceAgentAction,
    ResourceAgentMetadata,
    ResourceAgentName,
    ResourceAgentParameter,
    StandardProviderTuple,
)



def has_applicable_environment(environment):
    """
    dict environment - very likely os.environ
    """
    return bool(
        all(
            key in environment
            for key in [
                "COMP_WORDS",
                "COMP_LENGTHS",
                "COMP_CWORD",
                "PCS_AUTO_COMPLETE",
            ]
        )
        and environment["PCS_AUTO_COMPLETE"].strip() not in ("0", "")
        and environment["COMP_CWORD"].isdigit()
    )


def make_suggestions(environment, suggestion_tree):
    """
    dict environment - very likely os.environ
    dict suggestion_tree - {'acl': {'role': {'create': ...}}}...
    """
    if not has_applicable_environment(environment):
        raise EnvironmentError("Environment is not completion read")

    try:
        typed_word_list = _split_words(
            environment["COMP_WORDS"],
            environment["COMP_LENGTHS"].split(" "),
        )
    except EnvironmentError:
        return ""

    return "\n".join(
        _find_suggestions(
            suggestion_tree, typed_word_list, int(environment["COMP_CWORD"])
        )
    )


def _split_words(joined_words, word_lengths):
    cursor_position = 0
    words_string_len = len(joined_words)
    word_list = []
    for length in word_lengths:
        if not length.isdigit():
            raise EnvironmentError(
                "Length of word '{0}' is not digit".format(length)
            )
        next_position = cursor_position + int(length)
        if next_position > words_string_len:
            raise EnvironmentError(
                "Expected lengths are bigger than word lengths"
            )
        if (
            next_position != words_string_len
            and not joined_words[next_position].isspace()
        ):
            raise EnvironmentError("Words separator is not expected space")

        word_list.append(joined_words[cursor_position:next_position])
        cursor_position = next_position + 1

    if words_string_len > next_position:
        raise EnvironmentError("Expected lengths are smaller then word lengths")

    return word_list


def _find_suggestions(suggestion_tree, typed_word_list, word_under_cursor_idx):
    if not 1 <= word_under_cursor_idx <= len(typed_word_list):
        return []

    if len(typed_word_list) == word_under_cursor_idx:
        # not started type the last word yet
        word_under_cursor = ""
    else:
        word_under_cursor = typed_word_list[word_under_cursor_idx]

    words_for_current_cursor_position = _get_subcommands(
        suggestion_tree, typed_word_list[1:word_under_cursor_idx]
    )

    return [
        word
        for word in words_for_current_cursor_position
        if word.startswith(word_under_cursor)
    ]


def _get_subcommands(suggestion_tree, previous_subcommand_list):
    subcommand_tree = suggestion_tree
    for subcommand in previous_subcommand_list:
        if subcommand not in subcommand_tree:
            return []
        subcommand_tree = subcommand_tree[subcommand]
    return sorted(list(subcommand_tree.keys()))



from pcs.common.services import errors

from pcs_test.tools.command_env.mock_service_manager import Call


class ServiceManagerConfig:
    def __init__(self, calls):
        self.__calls = calls

    def start(
        self,
        service,
        instance=None,
        failure_msg=None,
        name="services.start",
        before=None,
        instead=None,
    ):
        self.__calls.place(
            name,
            Call(
                "start",
                service=service,
                instance=instance,
                return_value=None,
                exception=(
                    errors.StartServiceError(
                        service, message=failure_msg, instance=instance
                    )
                    if failure_msg
                    else None
                ),
            ),
            before=before,
            instead=instead,
        )

    def stop(
        self,
        service,
        instance=None,
        failure_msg=None,
        name="services.stop",
        before=None,
        instead=None,
    ):
        self.__calls.place(
            name,
            Call(
                "stop",
                service=service,
                instance=instance,
                return_value=None,
                exception=(
                    errors.StopServiceError(
                        service, message=failure_msg, instance=instance
                    )
                    if failure_msg
                    else None
                ),
            ),
            before=before,
            instead=instead,
        )

    def enable(
        self,
        service,
        instance=None,
        failure_msg=None,
        name="services.enable",
        before=None,
        instead=None,
    ):
        self.__calls.place(
            name,
            Call(
                "enable",
                service=service,
                instance=instance,
                return_value=None,
                exception=(
                    errors.EnableServiceError(
                        service, message=failure_msg, instance=instance
                    )
                    if failure_msg
                    else None
                ),
            ),
            before=before,
            instead=instead,
        )

    def disable(
        self,
        service,
        instance=None,
        failure_msg=None,
        name="services.disable",
        before=None,
        instead=None,
    ):
        self.__calls.place(
            name,
            Call(
                "disable",
                service=service,
                instance=instance,
                return_value=None,
                exception=(
                    errors.DisableServiceError(
                        service, message=failure_msg, instance=instance
                    )
                    if failure_msg
                    else None
                ),
            ),
            before=before,
            instead=instead,
        )

    def is_installed(
        self,
        service,
        return_value=True,
        name="services.is_installed",
        before=None,
        instead=None,
    ):
        self.__calls.place(
            name,
            Call(
                "is_installed",
                service=service,
                return_value=return_value,
            ),
            before=before,
            instead=instead,
        )

    def is_enabled(
        self,
        service,
        instance=None,
        return_value=True,
        name="services.is_enabled",
        before=None,
        instead=None,
    ):
        self.__calls.place(
            name,
            Call(
                "is_enabled",
                service=service,
                instance=instance,
                return_value=return_value,
            ),
            before=before,
            instead=instead,
        )

    def is_running(
        self,
        service,
        instance=None,
        name="services.is_running",
        return_value=True,
        before=None,
        instead=None,
    ):
        self.__calls.place(
            name,
            Call(
                "is_running",
                service,
                instance=instance,
                return_value=return_value,
            ),
            before=before,
            instead=instead,
        )

    def get_available_services(
        self,
        services,
        name="services.get_available_services",
        before=None,
        instead=None,
    ):
        self.__calls.place(
            name,
            Call("get_available_services", return_value=services),
            before=before,
            instead=instead,
        )



from unittest import (
    TestCase,
    mock,
)

from pcs.common import const
from pcs.common.reports import ReportItemSeverity as severities
from pcs.common.reports import codes as report_codes
from pcs.lib.commands.constraint import ticket as ticket_command

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import create_patcher

patch_commands = create_patcher("pcs.lib.commands.constraint.ticket")


class CreateTest(TestCase):
    def test_success_create(self):
        env_assist, config = get_env_tools(test_case=self)
        (
            config.runner.cib.load(
                filename="cib-empty-3.7.xml",
                resources="""
                    <resources>
                        <primitive id="resourceA" class="service" type="exim"/>
                    </resources>
                """,
            ).env.push_cib(
                optional_in_conf="""
                    <constraints>
                        <rsc_ticket
                            id="ticket-ticketA-resourceA-{role}"
                            rsc="resourceA"
                            rsc-role="{role}"
                            ticket="ticketA"
                            loss-policy="fence"
                        />
                    </constraints>
                """.format(
                    role=const.PCMK_ROLE_PROMOTED_PRIMARY
                )
            )
        )
        role = str(const.PCMK_ROLE_PROMOTED_LEGACY).lower()

        ticket_command.create(
            env_assist.get_env(),
            "ticketA",
            "resourceA",
            {
                "loss-policy": "fence",
                "rsc-role": role,
            },
        )
        env_assist.assert_reports(
            [
                fixture.deprecation(
                    report_codes.DEPRECATED_OPTION_VALUE,
                    option_name="role",
                    deprecated_value=role,
                    replaced_by=const.PCMK_ROLE_PROMOTED,
                )
            ]
        )

    def test_refuse_for_nonexisting_resource(self):
        env_assist, config = get_env_tools(test_case=self)
        config.runner.cib.load()
        env_assist.assert_raise_library_error(
            lambda: ticket_command.create(
                env_assist.get_env(),
                "ticketA",
                "resourceA",
                str(const.PCMK_ROLE_UNPROMOTED_LEGACY).lower(),
                {"loss-policy": "fence"},
            ),
            [
                (
                    severities.ERROR,
                    report_codes.ID_NOT_FOUND,
                    {
                        "context_type": "cib",
                        "context_id": "",
                        "id": "resourceA",
                        "expected_types": [
                            "bundle",
                            "clone",
                            "group",
                            "master",
                            "primitive",
                        ],
                    },
                    None,
                ),
            ],
            expected_in_processor=False,
        )


@patch_commands("get_constraints", mock.Mock())
class RemoveTest(TestCase):
    @patch_commands("ticket.remove_plain", mock.Mock(return_value=1))
    @patch_commands(
        "ticket.remove_with_resource_set", mock.Mock(return_value=0)
    )
    def test_successfully_remove_plain(self):
        self.assertTrue(ticket_command.remove(mock.MagicMock(), "T", "R"))

    @patch_commands("ticket.remove_plain", mock.Mock(return_value=0))
    @patch_commands(
        "ticket.remove_with_resource_set", mock.Mock(return_value=1)
    )
    def test_successfully_remove_with_resource_set(self):
        self.assertTrue(ticket_command.remove(mock.MagicMock(), "T", "R"))

    @patch_commands("ticket.remove_plain", mock.Mock(return_value=0))
    @patch_commands(
        "ticket.remove_with_resource_set", mock.Mock(return_value=0)
    )
    def test_raises_library_error_when_no_matching_constraint_found(self):
        self.assertFalse(ticket_command.remove(mock.MagicMock(), "T", "R"))
